/*
 sicmenu v0.61
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define CART_MAXSIZE 512*1024                          

#define ATR_SIZE (720*1024 + 16)

unsigned char data[CART_MAXSIZE];

int Map_L(int x)
{
  return ((x % 16384) >> 8) & 0xff;
}

int Map_H(int x)
{
  return x / 16384;
}

/*
 loaddata
*/
int loaddata(int p, char *fn)
{
  FILE *fp;
  int c;

  if (fp = fopen(fn, "rb"))
  {
    while ( (c = fgetc(fp)) != EOF )
      data[p++] = c;

    fclose(fp);
  }
  else
  {
    printf("could not open: '%s'\n", fn);
    exit(1);
  }

  return p;
}

int loaddata2(int p, char *fn)
{
  int p2 = loaddata(p, fn);
  data[p2++] = 0xfe;
  data[p2++] = 0xff;

  if ( (p2 & 0xff) != 0 )
    p2 = ((p2 / 256) * 256) + 256;

  return p2;
}

/*
 SetPlayer
*/
int SetPlayer(int p, char *fn, int initad, int acc, int x, int y, int playad)
{
  data[0x02] = Map_L(p);
  data[0x03] = Map_H(p);
  p = loaddata2(p, fn);

  data[0x06] = initad & 0xff; // init adr LO
  data[0x07] = (initad >> 8); // init adr HI
  data[0x08] = acc; // init ACC
  data[0x09] = x; // init X
  data[0x0a] = y; // init Y

  data[0x0b] = playad & 0xff; // player adr LO
  data[0x0c] = (playad >> 8); // player adr HI

  return p;
}


/* read16-bit word from a file (made to avoid endian problems */
unsigned short read_word(FILE* fp)
{
	unsigned short res = 0;
	unsigned char c;

	c = fgetc(fp);
	res = c;
	c = fgetc(fp);
	res |= c<<8;

	return res;
}

/*
 main
*/
int main(int argc, char **argv)
{
  FILE *fp;
  char fn[512], title[256];
  char fnout[512];
  int c;
  int i, j, k;
  int logo_maxh, logo_srcy;
  int size;
  int p, p2;
  int h;
  int margin_bot;
  char msg[1024];
  int msg_x, msg_y;
  char atr[ATR_SIZE], atr_fn[256];
  int filesize = 0;

  atr_fn[0] = 0;

  size = 0; // 1024 * 512;

  for (i = 0; i < CART_MAXSIZE; i++)
    data[i] = 0;

  data[0x1f] = 0xe0; // font address

  data[0x20] = 0x0c; // colpf1 (menu)
  data[0x21] = 0x00; // colpf2 (menu)
  data[0x22] = 0x00; // colpf3 (menu)
  data[0x23] = 0x00; // colpf0 (menu)
  data[0x24] = 0x00; // colbak (menu)

  data[0x25] = 0x04; // colpf0 (logo)
  data[0x26] = 0x08; // colpf1 (logo)
  data[0x27] = 0x0c; // colpf2 (logo)

  data[0x28] = 0x02; // graphics mode
  data[0x29] = 0x02; // left margin
  data[0x2a] = 0x01; // top margin
  margin_bot = 1;

  data[0x10] = 24; // number of screen lines

  msg_x = 0;
  msg_y = 0;
  msg[0] = 0;

  loaddata(0x2000, "data/sicmenu.obx");
  p = 0x4000;

  logo_maxh = 255;
  logo_srcy = 0;
  fn[0] = 0;

  strcpy(fnout, "cart.bin");

  for (i = 1; i < argc; i++)
  {
    if (strcmp(argv[i], "-logo_height") == 0)
    {
      logo_maxh = atoi(argv[++i]);  
    }
    else if (strcmp(argv[i], "-colors") == 0)
    {
      for (j = 0; j < 5; j++)
        if (i + 1 < argc && argv[i + 1][0] >= '0' && argv[i + 1][0] <= '9')
          data[0x20 + j] = atoi(argv[++i]);
    }
    else if (strcmp(argv[i], "-logo_colors") == 0)
    {
      for (j = 0; j < 3; j++)
        if (i + 1 < argc && argv[i + 1][0] >= '0' && argv[i + 1][0] <= '9')
          data[0x25 + j] = atoi(argv[++i]);
    }
    else if (strcmp(argv[i], "-logo_src_y") == 0)
    {
      logo_srcy = atoi(argv[++i]);
    }
    else if (strcmp(argv[i], "-font") == 0)
    {
      if ( strlen(argv[i + 1]) > 0 && (fp = fopen(argv[i + 1], "rb")) )
      {
        for (j = 0; j < 1024; j++)
          data[0x1c00 + j] = fgetc(fp);
     
        data[0x1f] = 0x9c;

        fclose(fp);

        i++;
      }
    }
    else if (strcmp(argv[i], "-screen") == 0)
    {
      if ( strlen(argv[i + 1]) > 0 && (fp = fopen(argv[i + 1], "rb")) )
      {
        for (j = 0; j < 1024; j++)
        {
          c = fgetc(fp);
          if (feof(fp))
            break;

          data[0x1800 + j] = c;
        }

        fclose(fp);

        i++;
      }
    }    
    else if (strcmp(argv[i], "-size") == 0)
    {
      size = atoi(argv[++i]) * 1024;
    }
    else if (strcmp(argv[i], "-intro") == 0)
    {
      data[0x00] = Map_L(p);
      data[0x01] = Map_H(p);
      p = loaddata2(p, argv[++i]);
    }
    else if (strcmp(argv[i], "-music") == 0)
    {
      data[0x04] = Map_L(p);
      data[0x05] = Map_H(p);
      j = p + 6;
      p = loaddata2(p, argv[++i]);

      if (data[j] == 'R' && data[j + 1] == 'M' && data[j + 2] == 'T')
      {
        p = SetPlayer(p, "data/rmtplayr.obx", 0x3400, 0x00, 0x00, 0x40, 0x3403);
      }
      else if (strcmp(argv[i] + strlen(argv[i]) - 3, "cmc") == 0 ||
               strcmp(argv[i] + strlen(argv[i]) - 3, "CMC") == 0 ||
               strcmp(argv[i] + strlen(argv[i]) - 3, "Cmc") == 0
               /* data[j - 4] == 0x00 && data[j - 3] == 0x84 */)
      {
        c = 0x40 - data[j - 3];
        printf("C=%i D=%i\n", c, data[j - 3]);
        data[j - 3] += c;
        data[j - 1] += c;
        for (k = j + 0x54; k < (j + 0x94); k++)
        {
           if (data[k] != 0xff)
             data[k] += c;
        }

        p = SetPlayer(p, "data/cmcplay.obx", 0x7803, 0x70, 0x00, 0x40, 0x7806);
      }
      else if (data[j - 4] == 0x00 && data[j - 3] == 0x50)
      {
        p = SetPlayer(p, "data/mptplay.obx", 0x4000, 0x00, 0x50, 0x00, 0x4003);
      }
      else
      {
        p = SetPlayer(p, "data/tmcplay.obx", 0x7800, 0x00, 0x28, 0x00, 0x7803);
      }        
    }
    else if (strcmp(argv[i], "-logo") == 0)
      i++;
    else if (strcmp(argv[i], "-mode") == 0)
    {
      data[0x28] = atoi(argv[++i]);
    }
    else if (strcmp(argv[i], "-margin") == 0)
    {
      data[0x29] = atoi(argv[++i]);
    }
    else if (strcmp(argv[i], "-margin_top") == 0)
    {
      data[0x2a] = atoi(argv[++i]);
    }
    else if (strcmp(argv[i], "-margin_bot") == 0)
    {
      margin_bot = atoi(argv[++i]);
    }
    else if (strcmp(argv[i], "-output") == 0 || strcmp(argv[i], "-o") == 0)
    {
      strcpy(fnout, argv[++i]);
    }
    else if (strcmp(argv[i], "-msg") == 0)
    {
      i++;
      for (j = 0; j < strlen(argv[i]); j++)
      {
        c = argv[i][j];
        if ( (c >= 'A' && c <= 'Z') || (c >= 0x5b && c <= 0x5f) || (c >= 0x20 && c <= 0x3f) )
          c -= 0x20;
        else if (c == '@')
          c = 0x20;

        msg[1 + j] = c;
      }
      msg[0] = j;
    }
    else if (strcmp(argv[i], "-msg_x") == 0)
    {
      msg_x = atoi(argv[++i]);
    }
    else if (strcmp(argv[i], "-msg_y") == 0)
    {
      msg_y = atoi(argv[++i]);
    }
    else if (strcmp(argv[i], "-atr") == 0)
    {
      strcpy(atr_fn, argv[++i]);
    }
    else
      strcpy(fn, argv[i]);
  }

  for (i = 0; i < msg[0]; i++)
    data[0x1800 + msg_x + (msg_y*40) + i] = msg[1 + i];

  for (i = 1; i < argc; i++)
  {   
    if (strcmp(argv[i], "-logo") == 0)
    {
      if (fp = fopen(argv[++i], "rb"))
      {
        p2 = p;
        p += 6;
        h = 0;

        for (j = 0; j < (40 * logo_srcy); j++)
          fgetc(fp);

        while (!feof(fp))
        {      
          for (j = 0; j < 40; j++)
            data[p++] = fgetc(fp);

          h++;

          if (h >= 102 || h >= logo_maxh)
            break;
        }

        p -= 40;
        h--;

        fclose(fp);
      }
      else
      {
        printf("could not open: '%s'\n", argv[i]);
        return -2;
      }

      data[p2] = 0xff;
      data[p2 + 1] = 0xff;
      data[p2 + 2] = 0x00;
      data[p2 + 3] = 0x10;
      data[p2 + 4] = (0x0fff + (h * 40) ) & 0xff;
      data[p2 + 5] = (( 0x0fff + (h * 40) ) >> 8) & 0xff;
      data[p++] = 0xfe;
      data[p++] = 0xff;

      if ( (p & 0xff) != 0 )
        p = ((p / 256) * 256) + 256;

      data[0x10] = (192 - h) / 8;
      data[0x11] = h;
      data[0x12] = Map_L(p2);
      data[0x13] = Map_H(p2);

      data[0x18] = ((192 - h) / 8) - data[0x2a];
    }
  }

  data[0x18] = data[0x10] - data[0x2a] - margin_bot; // number files per page

  if (strlen(fn) == 0)
  {
    printf("\nSIC!Menu v0.61 by pr0be, trub\n\n");
    printf("usage: %s xexlist [options]\n\n", argv[0]);
    printf("xexlist file:\n \"file1.xex\" \"title\"\n \"file2.xex\" \"title\"\n ...\n\n");
    printf("options:\n -output outputfile\n -size cartsize_in_kb\n -font fntfile\n -colors [colpf1] [colpf2] [colpf3] [colpf0] [colbak]\n");
    printf(" -margin left_margin\n -margin_top top_margin\n -margin_bot bottom_margin\n -mode 2 | 4\n -screen scrfile\n");
    printf(" -msg \"Message\"\n -msg_x msg_x_pos\n -msg_y msg_y_pos\n");
    printf(" -intro xexfile\n -music rmtfile | tmcfile | mptfile | cmcfile\n");
    printf(" -logo micfile\n -logo_colors [colpf0] [colpf1] [colpf2]\n -logo_height logoheight\n -logo_src_y y\n");
    printf(" -atr atrfile\n");
    printf("\n");

    return -1;
  }

  if (!(fp = fopen(fn, "rb")))
  {
    printf("could not open: '%s'\n", fn);
    return -2;
  }

  p2 = 0x40;
  while (!feof(fp))
  {
    while ( (c = fgetc(fp)) != '"' && !feof(fp));

    i = 0;
    while ( (c = fgetc(fp)) != '"' && !feof(fp) )
      fn[i++] = c;

    fn[i] = 0;

    while ( (c = fgetc(fp)) != '"' && !feof(fp));

    i = 0;
    while ( (c = fgetc(fp)) != '"' && !feof(fp) )
    {
      if ( (c >= 'A' && c <= 'Z') || (c >= 0x5b && c <= 0x5f) || (c >= 0x20 && c <= 0x3f) )
        c -= 0x20;
      else if (c == '@')
        c = 0x20;

      title[i++] = c;

	  if (i>25) break;	// to leave the place for size
    }

    if (i == 0)
      continue;

    data[p2++] = Map_L(p);
    data[p2++] = Map_H(p);

	filesize = loaddata(p, fn);	// just to get filesize
	filesize -= p;

    p = loaddata2(p, fn);


    for (j = 0; j < i; j++)
      data[p2++] = title[j];

    data[p2++] = 0xff;

    p2 = 32 + ((p2 / 32) * 32);

    data[p2-3] = filesize & 0xff;	// put file size at the end
    data[p2-2] = (filesize & 0xFF00) >> 8;
    data[p2-1] = (filesize & 0xFF0000) >> 16;

    data[0x19]++;
  }

  if (size == 0)
  {
    for (i = 128 * 1024; i < CART_MAXSIZE; i *= 2)
      if (p < i)
        break;

    size = i;
  }

  if (size > CART_MAXSIZE)
    size = CART_MAXSIZE;

  if (p > size)
  {
    printf("error: cartridge memory limit excess\n");     
    return -3;
  }

  fp = fopen(fnout, "wb");
  for (i = 0; i < size; i++)
    fputc(data[i], fp);
  fclose(fp);

  printf("%i file", (p2 - 0x40) / 32);
  if ( ((p2 - 0x40) / 32) > 1 )
    printf("s");

  printf(", %ib/%ib (%i%% free)\n", p, size, 100 - ((p*100) / size) );

  /*
   make ATR
  */
  p = 0;
  if (strlen(atr_fn) > 0)
  {
    char atr_head[] = {0x96, 0x02, 0x00, 0xB4, 0x80, 0x00, 0x00, 0x00,
                       0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
	  
    char t[] = {0x63, 0x63, 0x10, 0x00, 0x02, 0x53, 0x49, 0x43,
                0x20, 0x20, 0x20, 0x20, 0x20, 0x42, 0x49, 0x4e};

    unsigned short fadr;

    for (i = 0; i < ATR_SIZE; i++)
      atr[i] = 0;

    for (i = 0; i < sizeof(atr_head); i++)
      atr[0x0000 + i] = atr_head[i];

    for (i = 0; i < 16; i++)
      atr[0xb410 + i] = t[i];

    atr[0xb390] = 3;	// MyDOS

    p = 0xff90;
    p2 = 0x201;
    c = 0;
    for (i = 0; i < size; i++)
    {
      atr[p++] = data[i];
      c++;
      if (c == 125)
      {
        atr[p++] = (p2 >> 8);
        atr[p++] = p2 & 0xff;
        atr[p++] = 0x7d;
        p2++;
        c = 0;
      }
    }

    p = 0x8d + (((p - 0x10) / 128) * 128);
    atr[p++] = 0x00;
    atr[p++] = 0x00;
    atr[p++] = c;

    atr[0xb412] = (p2 - 0x200) >> 8;	// set file size
    atr[0xb411] = (p2 - 0x200) & 0xFF;


    fp = fopen("data/flasher.bin", "rb");
	if (!fp)
	{
      printf("could not open: 'data/flasher.bin'\n", fn);
      return -4;
	}
	
	fadr = read_word(fp);		// binary header
	if (fadr != 0xFFFF)
	{
	  printf("error: incorrect file 'data/flasher.bin' !\n");     
	  fclose(fp);
	  return -4;
	}

	// create disk header
	fadr = read_word(fp);			// get load address (little endian assumed)
	atr[0x12] = (fadr - 6) & 0xFF;	// load 6 bytes lower to fit the header
	atr[0x13] = (fadr - 6) >> 8;
	atr[0x14] = fadr & 0xFF;	// run address
	atr[0x15] = fadr >> 8;
	fadr = read_word(fp);		// get end address (and ignore)

	fadr = 0;					// now a byte count
	p = 0x16;					// code offset

    while (1)
    {
      c = fgetc(fp);
      if (c == EOF)
        break;

      atr[p++] = c;
	  fadr++;
    }
    fclose(fp);

	while (fadr % 128) fadr++;	// align to 128 boundary (sector count)

	atr[0x10] = (fadr / 128) >> 8;
	atr[0x11] = (fadr / 128) & 0xFF;

    fp = fopen(atr_fn, "wb");
    for (i = 0; i < ATR_SIZE; i++)
      fputc(atr[i], fp);

    fclose(fp);
  }
  
  return 0;
}
